/*
$Id: ow_2804.c,v 1.19 2007/03/24 23:06:59 alfille Exp $
    OWFS -- One-Wire filesystem
    OWHTTPD -- One-Wire Web Server
    Written 2003 Paul H Alfille
	email: palfille@earthlink.net
	Released under the GPL
	See the header file: ow.h for full attribution
	1wire/iButton system from Dallas Semiconductor
*/

/* General Device File format:
    This device file corresponds to a specific 1wire/iButton chip type
	( or a closely related family of chips )

	The connection to the larger program is through the "device" data structure,
	  which must be declared in the acompanying header file.

	The device structure holds the
	  family code,
	  name,
	  number of properties,
	  list of property structures, called "filetype".

	Each filetype structure holds the
	  name,
	  estimated length (in bytes),
	  aggregate structure pointer,
	  data format,
	  read function,
	  write funtion,
	  generic data pointer

	The aggregate structure, is present for properties that several members
	(e.g. pages of memory or entries in a temperature log. It holds:
	  number of elements
	  whether the members are lettered or numbered
	  whether the elements are stored together and split, or separately and joined
*/

/* Changes
    7/2004 Extensive improvements based on input from Serg Oskin
*/

#include <config.h>
#include "owfs_config.h"
#include "ow_2804.h"

/* ------- Prototypes ----------- */

/* DS2406 switch */
READ_FUNCTION(FS_r_mem);
WRITE_FUNCTION(FS_w_mem);
READ_FUNCTION(FS_r_page);
WRITE_FUNCTION(FS_w_page);
READ_FUNCTION(FS_r_pio);
WRITE_FUNCTION(FS_w_pio);
READ_FUNCTION(FS_sense);
READ_FUNCTION(FS_r_latch);
WRITE_FUNCTION(FS_w_latch);
READ_FUNCTION(FS_r_s_alarm);
WRITE_FUNCTION(FS_w_s_alarm);
READ_FUNCTION(FS_power);
READ_FUNCTION(FS_r_por);
READ_FUNCTION(FS_polarity);
WRITE_FUNCTION(FS_w_por);

/* ------- Structures ----------- */

struct aggregate A2804 = { 2, ag_numbers, ag_aggregate, };
struct aggregate A2804p = { 17, ag_numbers, ag_separate, };
struct filetype DS28E04[] = {
	F_STANDARD,
  {"memory", 550, NULL, ft_binary, fc_stable, {o: FS_r_mem}, {o: FS_w_mem}, {v:NULL},},
  {"pages", 0, NULL, ft_subdir, fc_volatile, {v: NULL}, {v: NULL}, {v:NULL},},
  {"pages/page", 32, &A2804p, ft_binary, fc_stable, {o: FS_r_page}, {o: FS_w_page}, {v:NULL},},
  {"polarity", 1, NULL, ft_yesno, fc_volatile, {o: FS_polarity}, {v: NULL}, {v:NULL},},
  {"power", 1, NULL, ft_yesno, fc_volatile, {o: FS_power}, {v: NULL}, {v:NULL},},
  {"por", 1, NULL, ft_yesno, fc_volatile, {o: FS_r_por}, {o: FS_w_por}, {v:NULL},},
  {"PIO", 1, &A2804, ft_bitfield, fc_stable, {o: FS_r_pio}, {o: FS_w_pio}, {v:NULL},},
  {"sensed", 1, &A2804, ft_bitfield, fc_volatile, {o: FS_sense}, {v: NULL}, {v:NULL},},
  {"latch", 1, &A2804, ft_bitfield, fc_volatile, {o: FS_r_latch}, {o: FS_w_latch}, {v:NULL},},
  {"set_alarm", 3, NULL, ft_unsigned, fc_stable, {o: FS_r_s_alarm}, {o: FS_w_s_alarm}, {v:NULL},},
};

DeviceEntryExtended(1C, DS28E04, DEV_alarm | DEV_resume | DEV_ovdr);

/* ------- Functions ------------ */

/* DS2804 */
static int OW_w_mem( BYTE * data,  size_t size,
					 off_t offset,  struct parsedname *pn);
static int OW_w_scratch( BYTE * data,  size_t size,
						 off_t offset,  struct parsedname *pn);
static int OW_w_pio( BYTE data,  struct parsedname *pn);
//static int OW_access( BYTE * data , const struct parsedname * pn ) ;
static int OW_clear( struct parsedname *pn);
static int OW_w_reg( BYTE * data,  size_t size,
					 off_t offset,  struct parsedname *pn);

/* 2804 memory read */
static int FS_r_mem(struct one_wire_query * owq)
{
	size_t pagesize = 32 ;
	/* read is not a "paged" endeavor, the CRC comes after a full read */
    if (OW_r_mem_simple( owq, 0, pagesize ))
		return -EINVAL;
    return OWQ_size(owq);
}

/* Note, it's EPROM -- write once */
static int FS_w_mem(struct one_wire_query * owq)
{
	size_t pagesize = 32 ;
	/* write is "byte at a time" -- not paged */
    if (OW_readwrite_paged(owq, 0, pagesize, OW_w_mem))
		return -EINVAL;
	return 0;
}

/* 2804 memory write */
static int FS_r_page(struct one_wire_query * owq)
{
	size_t pagesize = 32 ;
    if (OW_r_mem_simple( owq, OWQ_pn(owq).extension, pagesize))
		return -EINVAL;
    return OWQ_size(owq);
}

static int FS_w_page(struct one_wire_query * owq)
{
	size_t pagesize = 32 ;
	/* write is "byte at a time" -- not paged */
    if (OW_readwrite_paged(owq, OWQ_pn(owq).extension, pagesize, OW_w_mem))
		return -EINVAL;
	return 0;
}

/* 2804 switch */
static int FS_r_pio(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_pio ) ;
    OWQ_create_temporary( owq_pio, (char *) &data, 1, 0x0221, PN(owq) ) ;
    if (OW_r_mem_simple( owq_pio, 0, 0 ))
		return -EINVAL;
    OWQ_U(owq) = (data ^ 0xFF) & 0x03;	/* reverse bits */
	return 0;
}

/* write 2804 switch -- 2 values*/
static int FS_w_pio(struct one_wire_query * owq)
{
	BYTE data = 0;
	/* reverse bits */
    data = (OWQ_U(owq) ^ 0xFF) | 0xFC;	/* Set bits 2-7 to "1" */
    if (OW_w_pio(data, PN(owq)))
		return -EINVAL;
	return 0;
}

/* 2804 switch -- is Vcc powered?*/
static int FS_power(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_power ) ;
    OWQ_create_temporary( owq_power, (char *) &data, 1, 0x0225, PN(owq) ) ;
    if (OW_r_mem_simple( owq_power, 0, 0 ))
		return -EINVAL;
    OWQ_Y(owq) = UT_getbit(&data, 7);
	return 0;
}

/* 2904 switch -- power-on status of polrity pin */
static int FS_polarity(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_polarity ) ;
    OWQ_create_temporary( owq_polarity, (char *) &data, 1, 0x0225, PN(owq) ) ;
    if (OW_r_mem_simple( owq_polarity, 0, 0 ))
		return -EINVAL;
    OWQ_Y(owq) = UT_getbit(&data, 6);
	return 0;
}

/* 2804 switch -- power-on status of polrity pin */
static int FS_r_por(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_por ) ;
    OWQ_create_temporary( owq_por, (char *) &data, 1, 0x0225, PN(owq) ) ;
    if (OW_r_mem_simple( owq_por, 0, 0 ))
		return -EINVAL;
    OWQ_Y(owq) = UT_getbit(&data, 3);
	return 0;
}

/* 2804 switch -- power-on status of polrity pin */
static int FS_w_por(struct one_wire_query * owq)
{
	BYTE data;
    struct parsedname * pn = PN(owq) ;
	OWQ_make( owq_por ) ;
    OWQ_create_temporary( owq_por, (char *) &data, 1, 0x0225, pn ) ;
    if (OW_r_mem_simple( owq_por, 0, 0 ))
		return -EINVAL;			/* get current register */
	if (UT_getbit(&data, 3)) {	/* needs resetting? bit3==1 */
		data ^= 0x08;			/* flip bit 3 */
        if (OW_w_reg(&data, 1, 0x0225, pn))
			return -EINVAL;		/* reset */
		if (FS_r_por(owq))
			return -EINVAL;		/* reread */
        if (OWQ_Y(owq))
			return -EINVAL;		/* not reset despite our try */
	}
	return 0;					/* good */
}

/* 2804 switch PIO sensed*/
static int FS_sense(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_sense ) ;
    OWQ_create_temporary( owq_sense, (char *) &data, 1, 0x0220, PN(owq) ) ;
    if (OW_r_mem_simple( owq_sense, 0, 0 ))
		return -EINVAL;
    OWQ_U(owq) = (data) & 0x03;
	return 0;
}

/* 2804 switch activity latch*/
static int FS_r_latch(struct one_wire_query * owq)
{
	BYTE data;
	OWQ_make( owq_latch ) ;
    OWQ_create_temporary( owq_latch, (char *) &data, 1, 0x0222, PN(owq) ) ;
    if (OW_r_mem_simple( owq_latch, 0, 0 ))
		return -EINVAL;
    OWQ_U(owq) = data & 0x03;
	return 0;
}

/* 2804 switch activity latch*/
static int FS_w_latch(struct one_wire_query * owq)
{
    if (OW_clear(PN(owq)))
		return -EINVAL;
	return 0;
}

/* 2804 alarm settings*/
static int FS_r_s_alarm(struct one_wire_query * owq)
{
	BYTE data[3];
	OWQ_make( owq_alarm ) ;
    OWQ_create_temporary( owq_alarm, (char *) data, 3, 0x0223, PN(owq) ) ;
    if (OW_r_mem_simple( owq_alarm, 0, 0 ))
		return -EINVAL;
    OWQ_U(owq) = (data[2] & 0x03) * 100;
    OWQ_U(owq) += UT_getbit(&data[1], 0) | (UT_getbit(&data[0], 0) << 1);
    OWQ_U(owq) += UT_getbit(&data[1], 1) | (UT_getbit(&data[0], 1) << 1) * 10;
	return 0;
}

/* 2804 alarm settings*/
static int FS_w_s_alarm(struct one_wire_query * owq)
{
	BYTE data[3] = { 0, 0, 0, };
    UINT U = OWQ_U(owq) ;
	OWQ_make( owq_alarm ) ;
    OWQ_create_temporary( owq_alarm, (char *) &data[2], 1, 0x0225, PN(owq) ) ;
    if (OW_r_mem_simple( owq_alarm, 0, 0 ))
		return -EINVAL;
    data[2] |= (U / 100 % 10) & 0x03;
	UT_setbit(&data[1], 0, (int) (U % 10) & 0x01);
	UT_setbit(&data[1], 1, (int) (U / 10 % 10) & 0x01);
	UT_setbit(&data[0], 0, ((int) (U % 10) & 0x02) >> 1);
	UT_setbit(&data[0], 1, ((int) (U / 10 % 10) & 0x02) >> 1);
    if (OW_w_reg(data, 3, 0x0223, PN(owq)))
		return -EINVAL;
	return 0;
}

static int OW_w_scratch( BYTE * data,  size_t size,
						 off_t offset,  struct parsedname *pn)
{
	BYTE p[3 + 32 + 2] = { 0x0F, offset & 0xFF, (offset >> 8) & 0xFF, };
	struct transaction_log t[] = {
		TRXN_START,
		{p, NULL, 3 + size, trxn_match,},
		TRXN_END,
	};
	struct transaction_log tcrc[] = {
		TRXN_START,
		{p, NULL, 3 + size, trxn_match,},
		{&p[3 + size], NULL, 2, trxn_read,},
		{p, NULL, 3 + size + 2, trxn_crc16},
		TRXN_END,
	};

	memcpy(&p[3], data, size);
	if (((size + offset) & 0x1F) == 0) {	/* Check CRC if write is to end of page */
		return BUS_transaction(tcrc, pn);
	} else {
		return BUS_transaction(t, pn);
	}
}

/* pre-paged */
static int OW_w_mem( BYTE * data,  size_t size,
					 off_t offset,  struct parsedname *pn)
{
	BYTE p[4 + 32 + 2] = { 0xAA, offset & 0xFF, (offset >> 8) & 0xFF, };
	struct transaction_log tread[] = {
		TRXN_START,
		{p, NULL, 1, trxn_match,},
		{&p[1], NULL, 1 + size + 2, trxn_read,},
		{p, NULL, 4 + size + 2, trxn_crc16,},
		{data, &p[4], size, trxn_match,},
		TRXN_END,
	};
	struct transaction_log tcopy[] = {
		TRXN_START,
		{p, NULL, 3, trxn_match,},
		{&p[3], &p[3], 10, trxn_power,},
		TRXN_END,
	};

	if (OW_w_scratch(data, size, offset, pn)
		|| BUS_transaction(tread, pn)
		)
		return 1;
	p[0] = 0x55;
	return BUS_transaction(tcopy, pn);
}

//* write status byte */
static int OW_w_reg( BYTE * data,  size_t size,
					 off_t offset,  struct parsedname *pn)
{
	BYTE p[3] = { 0xCC, offset & 0xFF, (offset >> 8) & 0xFF, };
	struct transaction_log t[] = {
		TRXN_START,
		{p, NULL, 3, trxn_match,},
		{data, NULL, size, trxn_match,},
		TRXN_END,
	};
	return BUS_transaction(t, pn);
}

/* set PIO state bits: bit0=A bit1=B, value: open=1 closed=0 */
static int OW_w_pio( BYTE data,  struct parsedname *pn)
{
	BYTE p[3] = { 0x5A, data & 0xFF, (data & 0xFF) ^ 0xFF, };
	BYTE resp[1];
	struct transaction_log t[] = {
		TRXN_START,
		{p, NULL, 3, trxn_match,},
		{NULL, resp, 1, trxn_read,},
		TRXN_END,
	};

	return BUS_transaction(t, pn) || resp[0] != 0xAA;
}

/* Clear latches */
static int OW_clear( struct parsedname *pn)
{
	BYTE p[2] = { 0xC3, 0xFF };
	struct transaction_log t[] = {
		TRXN_START,
		{p, p, 2, trxn_read,},
		TRXN_END,
	};

	return BUS_transaction(t, pn) || p[0] != 0xC3 || p[1] != 0xAA;
}
